align 16
MergeSort25  PROC uses esi edi ebx Arr, count , stype , Scratch 
    ; MergeSort by Jim Giordano
    ; This routine sorts by merging streams of already sorted numbers.  It
    ;  differs from a classic merge sort in that the streams will vary in size
    ;  depending upon the input data, and a table of stream information is kept
    ;  to aid rapid merging.

    ; Arr is the starting address of the input array.  Arr can contain either an
    ;  array of values, or an array of addresses of the items to sort depending
    ;  upon stype.

    ; count is the number of items to sort
    
    ; stype is the type of input that Arr contains, and how they are to be sorted:

    ;   1 = addresses of strings to be sorted ascendingly
    ;   2 = addresses of strings, sort decending
    ;   3 = 32 bit signed integers, sort ascending
    ;   4 = 32 bit signed integers, sort decending
    ;   5 = addresses of 32 bit integers, sort ascending
    ;   6 = addresses of 32 bit integers, sort decending
    
    ; Scratch is the first address of a scratch buffer of at least count dwords
    ;  in size, or zero if the routine is to allocate the scratch buffer. 

    maxstream=4 ; 5  6  7  8; maximum number of streams being merged each level.
    maxlevel=15 ;13 12 11 10; worst case maximum number of levels needed to sort
                    ; maxstream=4 is faster on Intel, 6 on AMD but close enough.
    isize = 12  ; size of each mrg info block in bytes
    local rcount,level,heapflag,heaph
    local StreamStart,esiptr,firststream,ediptr
    local mrgstrt,mrgend,nexta,nextx,highestlevel
    local first[maxlevel]   ; address within mrg array of start of each level
    local next[maxlevel]    ; mrg array address of next free spot for each level
    local last[maxlevel]    ; last possible spot in each level
    local mrg[maxlevel*maxstream*isize/4]  ; info for mergestreams routine

    ; where L=level (up to maxlevel), n=stream number (up to maxstream)
    ; mrg(L,n,0) = [ebx+0] = actual address of current lowest value
    ; mrg(L,n,1) = [ebx+4] = address within array of the item under test
    ; mrg(L,n,2) = [ebx+8] = one past address of the last item in this stream

    .data
    MergeSortSetupFlag dd 0
    .code
    .if MergeSortSetupFlag==0
        call SetProtect ; allow code modification
        inc MergeSortSetupFlag
    .endif
    jmp SetupRoutine

    ; these are the various comparison routines
    ; entry to these routines will be
    ;    eax=address or value of the lowest (or highest for decending) item found so far
    ;    esi=address of the (value or address) of the item to compare
    ; return will be flags set from cmp
    ; may use regs edx and edi for scratch

    maxsize=28 ; must be a multiple of 4 and at least size of largest routine
    curmax=0

DoInt:          ; signed 32 bit integers
    chkstart=$
    mov edx,[eax]
    mov edi,[esi]
    cmp [edi],edx
    jmp chkstart+maxsize
    chksize=$-chkstart
    %if chksize gt curmax
        curmax=chksize
    endif

DoIntRev:      ; signed 32 bit integers, decending
    chkstart=$
    mov edx,[eax]
    mov edi,[esi]
    cmp edx,[edi]
    jmp chkstart+maxsize
    chksize=$-chkstart
    %if chksize gt curmax
        curmax=chksize
    endif

DoString:       ; addresses of strings, sort ascending
    chkstart=$
    mov edi,-1
    push esi    ; need another register for now
    mov esi,[esi]
    .repeat
        add edi,1
        mov dl,[eax+edi]
        cmp [esi+edi],dl
        .break .if !zero?   ; if different, we are done
        test dl,dl
    .until zero?    ; until out of chars to test, strings are the same
    pop esi        
    jmp chkstart+maxsize
    chksize=$-chkstart
    if chksize gt curmax
        curmax=chksize
    endif

DoStringRev:       ; addresses of strings, sort decending
    chkstart=$
    mov edi,-1
    push esi    ; need another register for now
    mov esi,[esi]
    .repeat
        add edi,1
        mov dl,[eax+edi]
        cmp dl,[esi+edi]
        .break .if !zero?   ; if different, we are done
        test dl,dl
    .until zero?    ; until out of chars to test, strings are the same
    pop esi        
    jmp chkstart+maxsize
    chksize=$-chkstart
    if chksize gt curmax
        curmax=chksize
    endif

DoIntInt:   ; this one works by value, not address, ascending
    chkstart=$
    cmp [esi],eax 
    jmp chkstart+maxsize
    chksize=$-chkstart
    %if chksize gt curmax
        curmax=chksize
    endif

DoIntIntRev:   ; this one works by value, not address, decending
    chkstart=$
    cmp eax,[esi] 
    jmp chkstart+maxsize
    chksize=$-chkstart
    %if chksize gt curmax
        curmax=chksize
    endif

;  note: due to masm bug, can't really test size in code, have to do manually.
;    if curmax gt maxsize
;        temp textequ %curmax
;          echo
;        % echo ********** routine size is larger than maxsize, routine size = temp **********
;          echo
;        .err        
;    endif

.data           ;     1        2       3            4         5      6
MergeSortTable dd DoString,DoStringRev,DoIntInt,DoIntIntRev,DoInt,DoIntRev
.code
SetupRoutine:
    mov eax,stype
    sub eax,1
    mov ebx,[MergeSortTable+eax*4]

    mov esi,ebx
    mov ecx,maxsize/4
    mov edi,TestLoc1
    rep movsd
    mov esi,ebx
    mov ecx,maxsize/4
    mov edi,TestLoc2
    rep movsd
    mov esi,ebx
    mov ecx,maxsize/4
    mov edi,TestLoc3
    rep movsd
    mov esi,ebx
    mov ecx,maxsize/4
    mov edi,TestLoc4
    rep movsd


    mov eax,count
    cmp eax,2
    jb nosort
    mov rcount,eax  ; remaining values to test

    mov heapflag,0
    .if Scratch == 0
        inv GetProcessHeap
        mov heaph,eax
        mov eax,count
        shl eax, 2  ; compute bytes of scratch heap size
        inv HeapAlloc,heaph,0,eax
        mov Scratch,eax
        mov heapflag,1
    .endif
        
    ; set up addresses of arrays
    xor ecx,ecx             ; level number (index into array)
    mov highestlevel,ecx    ; preset highest level so far
    lea eax,mrg             ; start of stream locations
    .repeat
        mov first[ecx],eax  ; address of start of each level
        mov next[ecx],eax   ; next available spot
        add eax,(maxstream-1)*isize ; number of possible streams this level
        mov last[ecx],eax   ; last possible spot for this level
        add eax,isize
        add ecx,4           ; levels are incremented by 4 (bytes/dword)
    .until ecx == 4*maxlevel

    mov esi,Arr
    mov esiptr,esi  ; where to start reading values for merge to scratch heap
    mov nexta,esi   ; where to store when merging back onto Arr
    mov edi,Scratch
    mov nextx,edi   ; where to store when merging to the scratch buffer

    comment * The procedure -

    GetNewData:
      level=0
      put info for Streams into mrg array (upto maxstream or until end of data)
      SetupStreamMerge:
        setup pointers to buffers
        merge to either Scratch for even levels or back to Arr for odd levels
        level=level+1, update mrg(level) values for this new level
        if higher than higest level so far, set highest level
        if this new level is full, merge it (goto SetupStreamMerge)
        if there are values left in Arr (rcount>0), goto GetNewData
        we are now out of new data, merge remaining streams:
        if this is not the highest level, merge this one up (SetupStreamMerge)
        if there is more than one stream in this level, merge them
        if current data sits in Scratch, move to Arr
    *

    align 16
    GetNewData:  ; back to here to get new streams from Arr until end of Arr
        mov level,0         ; new streams are always level 0
        mov ebx,next[0]     ; next available spot to store info for stream found
        mov esi,esiptr      ; where new data starts

        align 4
        GetStream:          ; get a stream of data from Arr
            mov eax,[esi]   ; get first value of stream
            mov [ebx+4],esi ; save start of stream
            mov [ebx],eax   ; save value
            add esi,4       ; check next value
            sub rcount,1    ; count down available values
            jz DoneNewData  ; out of values, all done, ebx was not incremented
            align 4
        TestLoc1:
            ; comparison routine will be inserted here
            ; eg. cmp [esi],[eax]   ; check next value to see direction of sequence
            ; and will jump to here when done
            org $+maxsize

            jl DoReversed   ; possible reversed sequence
            align 4         ; this one should already have been aligned
        DoNormal:           ; this section marks out an increasing stream
            mov eax,[esi]   ; get new larger value
            add esi,4       ; address of next value to test
            sub rcount,1    ; count down available values
            jz DoneNewData  ; out of new values, go merge what we have
            align 4
        TestLoc2:
            ; comparison routine will be inserted here
            ; eg. cmp [esi],[eax]   ; check next value to see direction of sequence
            ; and will jump to here when done
            org $+maxsize

            jge DoNormal    ; ok, go save it and test another

            cmp ebx,last[0] ; have we filled up a level?
            je DoneNewData  ; all done getting new data for now
            mov [ebx+8],esi ; save stop point
            add ebx,isize   ; set up to save next info for next stream
            jmp GetStream   ; go back and get another stream to merge
            align 4
        DoReversed:         ; this section marks out a decreasing stream
            mov eax,[esi]   ; get new smaller value
            add esi,4
            sub rcount,1
            jz dorev        ; out of new values, go reverse stream
            align 4
        TestLoc3:
            ; comparison routine will be inserted here
            ; eg. cmp [esi],[eax]   ; check next value to see direction of sequence
            ; and will jump to here when done
            org $+maxsize

            jl  DoReversed  ; ok, go save and test another.  note: if equal,
            ;    have to start over to maintain original order (keep stable)

          dorev:            ; now reverse the values found
            mov [ebx],eax   ; save first value of new stream
            mov ecx,esi
            sub ecx,4       ; address of last value in this stream
            mov edi,[ebx+4] ; current start of stream
            .repeat
                mov eax,[edi]   ; swap stream end for end
                mov edx,[ecx]
                mov [ecx],eax
                mov [edi],edx
                add edi,4
                sub ecx,4
            .until edi>=ecx
            cmp rcount,0    ; if we are out of new values, go do the merge
            je DoneNewData

            cmp ebx,last[0] ; have we filled up a level?
            je DoneNewData  ; yes, time to merge
            mov [ebx+8],esi ; no, save the stop point = one past last value
            add ebx,isize   ; where info for next stream will be stored
            jmp GetStream   ; go get another stream to merge

            align 4
      DoneNewData:
        mov [ebx+8],esi     ; save ending point (one past end of last stream)
        add ebx,isize
        mov esiptr,esi      ; save next to test
        mov next[0],ebx     ; save next free cell for testing below

        ; Merge Streams
        align 4
      SetupStreamMerge:
        mov ecx,level       ; time to merge the streams found in this level
        mov edi,nextx       ; assume merging to heap
        test ecx,4          ; level increments by 4 for indexing into arrays
        cmovnz edi,nexta    ; nope, moving back to Arr
        mov StreamStart,edi ; save start of new stream
        mov ediptr,edi      ; save store location
        mov edx,first[ecx]  ; location of mrg info, first stream in this level
        mov esi,[edx+4]     ; location of first value in this stream
        mov firststream,esi ; save for wrapup

        mov eax,next[ecx]   ; one past last stream in this level
        sub eax,isize       ; start last stream in this level
        mov esi,first[ecx]  ; start of first stream in this level
        mov mrgstrt,esi     ; save begin  of stream in this level
        mov mrgend,eax      ; get last loc containing info for merge

        .if eax>esi         ; is there more than one stream to merge?
            ; eax=address of smallest value so far
            ; ebx=scratch for compare routine
            ; ecx=address of smallest stream info block so far
            ; edx=scratch for compare routine
            ; esi=pointer into stream
            ; edi=output location.  will be saved so edi can be compare scratch
            align 4
          mergestreams:     ; merge streams found in mrg info array
            mov esi,mrgstrt ; start of info for first stream in this level
            mov ecx,esi     ; save address of current smallest value
            mov eax,[esi]   ; first value in first string
            .repeat
                add esi,isize   ; compare to value in next stream
              TestLoc4:
                ;eg. cmp [esi],[eax]
                org $+maxsize
                cmovl eax,[esi] ; if smaller, save it
                cmovl ecx,esi   ; and also it's address
            .until esi==mrgend  ; esi will equal mrgend when we exit
            mov edi,ediptr
            mov [edi],eax       ; now have smallest of the group, save it
            mov edx,[ecx+4]     ; get address of smallest value
            add edi,4
            add edx,4           ; address of next value to test in this stream
            mov ediptr,edi      ; save for next time
            .if edx!=[ecx+8]    ; are we done with this stream?
                mov [ecx+4],edx ; no, save the new address
                mov eax,[edx]   ; get the value
                mov [ecx],eax   ; save new value
                jmp mergestreams    ; and go compare everyone again
            .endif
            ; we ran out of values in this stream, move last stream into this
            ;  spot if it's not the last one
            ; esi is now equal to the offset into mrg array of last stream used
            .if esi != ecx      ; was the last stream the smallest?
                mov eax,[esi]   ; no, mov last group into this used up position
                mov [ecx],eax
                mov eax,[esi+4]
                mov [ecx+4],eax
                mov eax,[esi+8]
                mov [ecx+8],eax
            .endif
            sub esi,isize       ; there is now one less stream to test
            mov mrgend,esi      ; save the address of the last stream this level
            cmp esi,mrgstrt     ; are we done?
            ja  mergestreams    ; no, go merge remaining streams

        .endif
        ; only one stream left, copy it to output
        mov edx,[esi+4] ; location of current value last remaining stream
        .repeat
            mov eax,[edx]   ; get value
            add edx,4       ; location of next value
            mov [edi],eax   ; store
            add edi,4       ; etc
        .until edx==[esi+8] ; is it past the last one?

        mov eax,firststream ; address of the first input stream
        mov ecx,level
        test ecx,4          ; are we copying to Arr or Scratch?
        .if zero?
            mov nextx,edi   ; update next store location
            .if eax<nexta   ; save the earliest freed up location
                mov nexta,eax
            .endif
        .else
            mov nexta,edi
            .if eax<nextx
                mov nextx,eax
            .endif
        .endif

        mov eax,mrgstrt
        mov next[ecx],eax   ; reset this level to zero streams
        add ecx,4           ; update next higher level info for streams merged
        mov level,ecx
        .if ecx>highestlevel
            mov highestlevel,ecx    ; save highest level for final merge
        .endif
        mov edx,next[ecx]   ; location of next store
        mov eax,StreamStart ; start of new stream just created
        mov [edx+4],eax
        mov [edx+8],edi     ; next output address = one past end of this stream
        mov eax,[eax]       ; copy the first value of the stream
        mov [edx],eax
        add edx,isize
        mov next[ecx],edx   ; where next new stream will be stored
        cmp edx,last[ecx]   ; if this level is now full, go merge these steams
        ja SetupStreamMerge
        cmp rcount,0        ; if there are values left to do, go do it
        jne GetNewData
        ; out of data, merge all leftover streams from here up to highest level
        cmp ecx,highestlevel    ; if not the highest level, merge it up
        jne SetupStreamMerge
        sub edx,isize       ; if more than one stream in this level, go merge
        cmp edx,first[ecx]
        ja  SetupStreamMerge
        test ecx,4          ; is final data in Scratch?
        jz completelydone   ; no, done
        mov esi,Scratch       ; yes, copy back to Arr
        mov edi,Arr
        mov ecx,count
        rep movsd

completelydone:
   .if heapflag
       inv HeapFree,heaph,0,Scratch
   .endif

nosort:
    ret

.data
smcerror db "MergeSort Setup Error",0
.code
SetProtect:
    mov ecx,nosort
    mov edx,GetNewData
    sub ecx,edx   ; amount to unprotect
    inv VirtualProtect,edx,ecx,PAGE_EXECUTE_READWRITE,addr rcount
    .if eax == 0
        inv MessageBox,0,addr smcerror,0,0
        jmp nosort
    .endif
    invoke GetCurrentProcess
    mov edx,GetNewData
    mov ecx,nosort
    sub ecx,edx
    invoke FlushInstructionCache,eax,edx,ecx
    retn

MergeSort25  ENDP
mrg25=$-MergeSort25
